import os
from pathlib import Path
import shutil
import logging

log = logging.getLogger(__name__)

class ProfileManager:
    def __init__(self):
        self.config_dir = Path.home() / ".config" / "mastui"
        self.profiles_dir = self.config_dir / "profiles"
        self.profiles_dir.mkdir(parents=True, exist_ok=True)
        self.last_profile_file = self.config_dir / "last_profile.txt"

    def get_last_profile(self) -> str | None:
        """Reads the name of the last used profile."""
        try:
            if self.last_profile_file.exists():
                return self.last_profile_file.read_text().strip()
        except (IOError, OSError) as e:
            log.error(f"Failed to read last profile file: {e}", exc_info=True)
            return None
        return None

    def set_last_profile(self, profile_name: str):
        """Saves the name of the last used profile."""
        try:
            self.last_profile_file.write_text(profile_name)
        except (IOError, OSError) as e:
            log.error(f"Failed to write last profile file: {e}", exc_info=True)


    def get_profiles(self) -> list[str]:
        """Returns a list of available profile names."""
        return [d.name for d in self.profiles_dir.iterdir() if d.is_dir()]

    def get_profile_path(self, profile_name: str) -> Path:
        """Returns the path to a specific profile directory."""
        return self.profiles_dir / profile_name

    def create_profile(self, profile_name: str, env_content: str):
        """Creates a new profile directory and .env file."""
        profile_path = self.get_profile_path(profile_name)
        profile_path.mkdir(exist_ok=True)
        (profile_path / ".env").write_text(env_content)

    def delete_profile(self, profile_name: str):
        """Deletes a profile directory."""
        profile_path = self.get_profile_path(profile_name)
        if profile_path.exists():
            shutil.rmtree(profile_path)

    def migrate_old_profile(self) -> bool:
        """Migrates the old single-profile setup to the new profiles directory."""
        old_env_file = self.profiles_dir.parent / ".env"
        old_cache_db = self.profiles_dir.parent / "cache.db"
        
        if old_env_file.exists():
            default_profile_path = self.get_profile_path("default")
            default_profile_path.mkdir(exist_ok=True)
            
            # Move the old .env file
            shutil.move(str(old_env_file), str(default_profile_path / ".env"))
            
            # Move the old cache.db if it exists
            if old_cache_db.exists():
                shutil.move(str(old_cache_db), str(default_profile_path / "cache.db"))
            
            return True
        return False

profile_manager = ProfileManager()



import sqlite3
import json
from pathlib import Path
import logging
from datetime import datetime, timezone, timedelta
import os

log = logging.getLogger(__name__)

class CustomJsonEncoder(json.JSONEncoder):
    """
    Custom JSON encoder to handle datetime objects.
    """
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)

class Cache:
    def __init__(self, db_path: Path):
        self.db_path = db_path
        self.initialize_database()

    def _get_conn(self):
        """Get a new database connection."""
        try:
            conn = sqlite3.connect(self.db_path)
            conn.row_factory = sqlite3.Row
            return conn
        except sqlite3.Error as e:
            log.error(f"Database connection failed: {e}", exc_info=True)
            return None

    def initialize_database(self):
        """Create tables if they don't exist. Called once at startup."""
        conn = self._get_conn()
        if not conn:
            return
        try:
            cursor = conn.cursor()
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS posts (
                    id TEXT PRIMARY KEY,
                    timeline_id TEXT NOT NULL,
                    created_at TEXT NOT NULL,
                    data TEXT NOT NULL
                )
            """)
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS conversations (
                    id TEXT PRIMARY KEY,
                    unread INTEGER NOT NULL,
                    last_status_id TEXT,
                    data TEXT NOT NULL
                )
            """)
            cursor.execute("CREATE INDEX IF NOT EXISTS idx_timeline_id ON posts (timeline_id, id)")
            conn.commit()
        except sqlite3.Error as e:
            log.error(f"Failed to create tables: {e}", exc_info=True)
        finally:
            if conn:
                conn.close()

    def get_conversations(self):
        """Get all conversations from the database."""
        conn = self._get_conn()
        if not conn:
            return []
        try:
            cursor = conn.cursor()
            cursor.execute("SELECT data FROM conversations")
            rows = cursor.fetchall()
            # Manually set 'unread' to a boolean, as it's stored as an integer
            conversations = []
            for row in rows:
                conv = json.loads(row['data'])
                conv['unread'] = bool(conv.get('unread'))
                # Parse created_at back to datetime for consistency
                if conv.get('last_status') and conv['last_status'].get('created_at'):
                    ts_str = conv['last_status']['created_at']
                    if isinstance(ts_str, str):
                        conv['last_status']['created_at'] = datetime.fromisoformat(ts_str)
                conversations.append(conv)
            return conversations
        except sqlite3.Error as e:
            log.error(f"Failed to get conversations: {e}", exc_info=True)
            return []
        finally:
            if conn:
                conn.close()

    def bulk_insert_conversations(self, conversations: list):
        """Bulk insert conversations into the database."""
        if not conversations:
            return
        conn = self._get_conn()
        if not conn:
            return
        try:
            cursor = conn.cursor()
            convos_to_insert = []
            for convo in conversations:
                last_status = convo.get('last_status')
                convos_to_insert.append((
                    convo['id'],
                    1 if convo.get('unread') else 0,
                    last_status['id'] if last_status else None,
                    json.dumps(convo, cls=CustomJsonEncoder)
                ))
            
            cursor.executemany(
                "INSERT OR REPLACE INTO conversations (id, unread, last_status_id, data) VALUES (?, ?, ?, ?)",
                convos_to_insert
            )
            conn.commit()
            log.info(f"Inserted/updated {len(convos_to_insert)} conversations")
        except sqlite3.Error as e:
            log.error(f"Failed to bulk insert conversations: {e}", exc_info=True)
        finally:
            if conn:
                conn.close()

    def mark_conversation_as_read(self, conversation_id: str):
        """Mark a conversation as read in the database."""
        conn = self._get_conn()
        if not conn:
            return
        try:
            cursor = conn.cursor()
            cursor.execute("UPDATE conversations SET unread = 0 WHERE id = ?", (conversation_id,))
            conn.commit()
            log.info(f"Marked conversation {conversation_id} as read in cache.")
        except sqlite3.Error as e:
            log.error(f"Failed to mark conversation as read: {e}", exc_info=True)
        finally:
            if conn:
                conn.close()

    def bulk_insert_posts(self, timeline_id: str, posts: list):
        """Bulk insert posts into the database."""
        if not posts:
            return
        conn = self._get_conn()
        if not conn:
            return
        try:
            cursor = conn.cursor()
            posts_to_insert = []
            for post in posts:
                created_at_str = (post.get('reblog') or post).get('created_at')
                if not created_at_str:
                    log.warning(f"Post {post.get('id')} has no created_at timestamp. Skipping.")
                    continue

                if isinstance(created_at_str, datetime):
                    created_at = created_at_str
                else:
                    created_at = datetime.fromisoformat(created_at_str.replace("Z", "+00:00"))
                
                posts_to_insert.append((
                    post['id'],
                    timeline_id,
                    created_at.isoformat(),
                    json.dumps(post, cls=CustomJsonEncoder)
                ))
            
            cursor.executemany(
                "INSERT OR REPLACE INTO posts (id, timeline_id, created_at, data) VALUES (?, ?, ?, ?)",
                posts_to_insert
            )
            conn.commit()
            log.info(f"Inserted/updated {len(posts_to_insert)} posts for timeline '{timeline_id}'")
        except sqlite3.Error as e:
            log.error(f"Failed to bulk insert posts: {e}", exc_info=True)
        finally:
            if conn:
                conn.close()

    def get_latest_post_timestamp(self, timeline_id: str) -> datetime | None:
        """Get the timestamp of the latest post in the cache for a timeline."""
        conn = self._get_conn()
        if not conn:
            return None
        try:
            cursor = conn.cursor()
            cursor.execute("SELECT MAX(created_at) FROM posts WHERE timeline_id = ?", (timeline_id,))
            row = cursor.fetchone()
            if row and row[0]:
                return datetime.fromisoformat(row[0])
            return None
        except sqlite3.Error as e:
            log.error(f"Failed to get latest post timestamp: {e}", exc_info=True)
            return None
        finally:
            if conn:
                conn.close()

    def get_posts(self, timeline_id: str, limit: int = 20, max_id: str = None):
        """Get posts from the database, ordered by ID."""
        conn = self._get_conn()
        if not conn:
            return []
        try:
            cursor = conn.cursor()
            query = "SELECT data FROM posts WHERE timeline_id = ?"
            params = [timeline_id]

            if max_id:
                query += " AND id < ?"
                params.append(max_id)

            query += " ORDER BY id DESC LIMIT ?"
            params.append(limit)

            cursor.execute(query, params)
            rows = cursor.fetchall()
            return [json.loads(row['data']) for row in rows]
        except sqlite3.Error as e:
            log.error(f"Failed to get posts: {e}", exc_info=True)
            return []
        finally:
            if conn:
                conn.close()

    def prune_image_cache(self, days: int = 30) -> int:
        """Prune the image cache of files older than a certain number of days."""
        count = 0
        cutoff = datetime.now(timezone.utc) - timedelta(days=days)
        
        image_cache_dir = self.db_path.parent / "image_cache"
        if not image_cache_dir.exists():
            return 0

        for filename in os.listdir(image_cache_dir):
            file_path = image_cache_dir / filename
            try:
                if file_path.is_file():
                    modified_time = datetime.fromtimestamp(file_path.stat().st_mtime, tz=timezone.utc)
                    if modified_time < cutoff:
                        file_path.unlink()
                        count += 1
            except Exception as e:
                log.error(f"Error pruning cache file {file_path}: {e}", exc_info=True)
        
        log.info(f"Pruned {count} items from the image cache.")
        return count


from textual.widgets import Header, Static

class CustomHeader(Header):
    """A custom header with a DM notification icon."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.tall = False

    def compose(self):
        yield from super().compose()
        yield Static("📩", id="dm_notification_icon", classes="hidden")

    def show_dm_notification(self):
        """Show the DM notification icon."""
        self.query_one("#dm_notification_icon").remove_class("hidden")

    def hide_dm_notification(self):
        """Hide the DM notification icon."""
        self.query_one("#dm_notification_icon").add_class("hidden")



from textual.screen import ModalScreen
from textual.widgets import Button, Select, Static
from textual.containers import Vertical
from mastui.logo import LogoWidget

class ProfileSelectionScreen(ModalScreen):
    """A modal screen for selecting a profile."""

    BINDINGS = [("escape", "dismiss", "Dismiss")]

    def __init__(self, profiles: list[str], **kwargs):
        super().__init__(**kwargs)
        self.profiles = profiles

    def compose(self):
        with Vertical(id="profile-selection-dialog"):
            yield LogoWidget()
            yield Static("Select a profile to continue:", classes="label")
            yield Select([(p, p) for p in self.profiles], id="profile-select")
            yield Button("Login", variant="primary", id="login-button")
            yield Button("Add New Profile", id="add-profile-button")

    def on_mount(self) -> None:
        # Auto-focus the select widget if there are profiles
        if self.profiles:
            self.query_one("#profile-select").focus()
        else:
            self.query_one("#add-profile-button").focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "login-button":
            selected_profile = self.query_one("#profile-select").value
            if selected_profile:
                self.dismiss(selected_profile)
        elif event.button.id == "add-profile-button":
            self.dismiss("add_new_profile")



import html2text
import re
from datetime import datetime, timezone
from dateutil.parser import parse
from bs4 import BeautifulSoup
import logging
import pprint

log = logging.getLogger(__name__)

LANGUAGE_OPTIONS = [
    ("Chinese", "zh"),
    ("Danish", "da"),
    ("English", "en"),
    ("French", "fr"),
    ("German", "de"),
    ("Japanese", "ja"),
    ("Korean", "ko"),
    ("Spanish", "es"),
]

VISIBILITY_OPTIONS = [
    ("🌐 Public", "public"),
    ("🔑 Unlisted", "unlisted"),
    ("👥 Followers-only", "private"),
    ("🔒 Direct", "direct"),
]

MARKDOWN_LINK_REGEX = re.compile(r"\\[^\\]+\\]\(([^)]+)\\\)")


def markdown_links_to_html(text: str) -> str:
    """Converts Markdown-style links in a string to HTML <a> tags."""
    if not text:
        return ""
    return MARKDOWN_LINK_REGEX.sub(
        lambda m: f'<a href="{m.group(2)}">{m.group(1)}</a>', text
    )


def to_markdown(html: str) -> str:
    """
    Converts mixed HTML/Markdown content to terminal-friendly markdown.
    """
    if not html:
        return ""

    # First, parse with BeautifulSoup to handle HTML and decode entities.
    soup = BeautifulSoup(html, "html.parser")

    # Convert the parsed soup back to a string.
    cleaned_html = str(soup)

    # Now, convert any remaining Markdown-style links to HTML <a> tags.
    html_with_links = markdown_links_to_html(cleaned_html)

    # Finally, convert the fully-HTML content to terminal-friendly text.
    h = html2text.HTML2Text()
    h.ignore_links = False
    h.body_width = 0  # Don't wrap lines
    return h.handle(html_with_links)


def get_full_content_md(status: dict) -> str:
    """Gets the full markdown content for a status, including media."""
    if not status:
        return ""

    html_content = status.get("content") or status.get("note") or ""

    content_md = to_markdown(html_content)

    if status.get("media_attachments"):
        media_infos = []
        for media in status["media_attachments"]:
            media_type = media.get("type", "media").capitalize()
            description = media.get("description")

            if description:
                media_infos.append(f"[{media_type} showing: {description}]")
            else:
                media_infos.append(f"[{media_type} attached]")

        if content_md.strip():
            content_md += "\n\n" + "\n".join(media_infos)
        else:
            content_md = "\n".join(media_infos)

    return content_md


def html_to_plain_text(html: str) -> str:
    """
    Converts HTML content from a status to plain text, handling mentions correctly.
    """
    if not html:
        return ""
    
    soup = BeautifulSoup(html, 'html.parser')
    
    # Find all mention links and replace them with @user@host text
    for mention in soup.find_all('a', class_='mention'):
        # The text content is usually just the username, but we can get the full acct from the href
        href = mention.get('href', '')
        try:
            # A bit of parsing to get the acct from a URL like https://instance.social/@username
            acct = f"@{href.split('/@')[-1]}"
            mention.replace_with(acct)
        except Exception:
            # Fallback to the text if parsing fails
            mention.replace_with(mention.get_text(strip=True))

    # Replace <p> tags with newlines for paragraph breaks
    for p in soup.find_all('p'):
        p.append('\n')

    # Get the rest of the text, stripping other HTML tags
    return soup.get_text().strip()


def format_datetime(dt_obj) -> str:
    """Formats a datetime string or object into YYYY-MM-DD HH:MM."""
    if isinstance(dt_obj, str):
        dt = parse(dt_obj)
    else:
        dt = dt_obj
    
    if dt.tzinfo:
        dt = dt.astimezone()
    
    return dt.strftime('%Y-%m-%d %H:%M')



import json
from pathlib import Path
import logging
from textual.theme import Theme

log = logging.getLogger(__name__)

# The path to the user's custom themes file
CUSTOM_THEMES_PATH = Path.home() / ".config" / "mastui" / "custom-themes.json"

# The default retro-green theme, to be used as a template
DEFAULT_THEME_DATA = {
    "name": "retro-green",
    "styles": {
        "app": "bg: #0F0F0F text: #A0A0A0",
        "header": "bg: #1A1A1A text: #E0E0E0",
        "footer": "bg: #1A1A1A text: #A0A0A0",
        "timeline_title": "text: #00FF00 bold",
        "timeline-item.selected": "bg: #1A1A1A",
        "timeline-item.favourited": "text: #FFFF00",
        "timeline-item.reblogged": "text: #00FF00",
        "post-footer": "text: #606060",
        "boost-header": "text: #00C000",
        "poll-header": "text: #00FF00",
        "poll-total-votes": "text: #A0A0A0",
        "poll-expiry": "text: #606060",
        "button": "bg: #2A2A2A text: #E0E0E0",
        "button.primary": "bg: #008000 text: #FFFFFF",
        "input": "bg: #1A1A1A text: #E0E0E0",
        "select": "bg: #1A1A1A text: #E0E0E0",
        "switch": "bg: #2A2A2A",
        "switch--on": "bg: #008000",
    },
    "palettes": {
        "primary": "#00FF00",
        "secondary": "#008000",
        "accent": "#00C000",
        "foreground": "#A0A0A0",
        "background": "#0F0F0F",
        "surface": "#1A1A1A",
        "panel": "#1A1A1A",
        "success": "#00FF00",
        "warning": "#FFFF00",
        "error": "#FF0000",
        "dark": True,
    },
}

def create_default_themes_file_if_not_exists():
    """
    Creates the custom-themes.json file with the default retro-green
    theme if the file does not already exist.
    """
    if CUSTOM_THEMES_PATH.exists():
        return

    log.info(f"Creating default custom themes file at {CUSTOM_THEMES_PATH}")
    try:
        CUSTOM_THEMES_PATH.parent.mkdir(parents=True, exist_ok=True)
        with open(CUSTOM_THEMES_PATH, "w") as f:
            # The file should contain a list of themes
            json.dump([DEFAULT_THEME_DATA], f, indent=4)
    except Exception as e:
        log.error(f"Failed to create custom themes file: {e}", exc_info=True)

def load_custom_themes() -> list[Theme]:
    """
    Loads themes from the custom-themes.json file.
    """
    # Ensure the file exists before trying to read it
    create_default_themes_file_if_not_exists()

    themes = []
    try:
        with open(CUSTOM_THEMES_PATH, "r") as f:
            themes_data = json.load(f)
        
        if not isinstance(themes_data, list):
            log.error("Custom themes file is not a valid list of themes.")
            return []

        for theme_data in themes_data:
            try:
                # Filter the palettes to only include keys the constructor expects
                valid_palette_keys = [
                    "primary", "secondary", "accent", "foreground", "background",
                    "success", "warning", "error", "surface", "panel", "dark"
                ]
                filtered_palette = {
                    key: theme_data["palettes"][key]
                    for key in valid_palette_keys
                    if key in theme_data["palettes"]
                }

                theme = Theme(
                    name=theme_data["name"],
                    **filtered_palette
                )
                theme.styles = theme_data["styles"]
                themes.append(theme)
                log.info(f"Successfully loaded custom theme: '{theme.name}'")
            except KeyError as e:
                log.error(f"Missing key in custom theme '{theme_data.get('name', 'N/A')}': {e}")
    except json.JSONDecodeError as e:
        log.error(f"Error decoding custom themes file: {e}")
    except Exception as e:
        log.error(f"Failed to load custom themes: {e}", exc_info=True)
    
    return themes



import html2text
from textual.widgets import (
    Static,
    LoadingIndicator,
    RadioSet,
    RadioButton,
    Button,
    Input,
    Markdown,
)
from textual.widget import Widget
from textual.containers import VerticalScroll, Vertical, Horizontal
from textual import events, on
from textual.message import Message
from mastui.utils import get_full_content_md, format_datetime, to_markdown
from mastui.reply import ReplyScreen
from mastui.image import ImageWidget
from mastui.messages import SelectPost, VoteOnPoll, ViewHashtag
import logging
from datetime import datetime

log = logging.getLogger(__name__)


class PostMessage(Message):
    """A message relating to a post."""

    def __init__(self, post_id: str) -> None:
        self.post_id = post_id
        super().__init__()


class LikePost(PostMessage):
    """A message to like a post."""

    def __init__(self, post_id: str, favourited: bool):
        super().__init__(post_id)
        self.favourited = favourited


class BoostPost(PostMessage):
    """A message to boost a post."""

    pass


class RemovePollChoice(Message):
    """A message to remove a poll choice."""

    def __init__(self, poll_choice_widget: Widget) -> None:
        self.poll_choice_widget = poll_choice_widget
        super().__init__()


class PollChoiceMounted(Message):
    """A message to indicate that a poll choice has been mounted."""

    pass


class PollChoice(Horizontal):
    """A widget for a single poll choice."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_class("poll-choice")

    def compose(self):
        yield Input(placeholder="Choice...")
        yield Button("X", variant="error", classes="remove-choice", disabled=True)

    def on_mount(self) -> None:
        """When the widget is mounted, tell the parent to check button states."""
        self.post_message(PollChoiceMounted())

    @on(Button.Pressed, ".remove-choice")
    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.post_message(RemovePollChoice(self))


class PollWidget(Vertical):
    """A widget to display a poll."""

    def __init__(self, poll: dict, timeline_id: str, post_id: str, **kwargs):
        super().__init__(**kwargs)
        self.poll = poll
        self.timeline_id = timeline_id
        self.post_id = post_id
        self.add_class("poll-container")
        self.styles.height = "auto"

    def compose(self):
        # The PollWidget itself acts as a vertical container for its children.
        if self.poll.get("voted") or self.poll.get("expired"):
            # Show results
            total_votes = self.poll.get("votes_count", 0)
            yield Static("Poll Results:", classes="poll-header")
            for i, option in enumerate(self.poll["options"]):
                votes = option.get("votes_count", 0)
                percentage = (votes / total_votes * 100) if total_votes > 0 else 0

                is_own_vote = i in self.poll.get("own_votes", [])

                label_prefix = "✓ " if is_own_vote else "  "
                label = f"{label_prefix}{option['title']} ({votes} votes, {percentage:.2f}%)"

                bar_color = "green" if is_own_vote else "blue"
                result_bar = Static(f"[{bar_color}] {'█' * int(percentage / 2)} [/]")

                yield Static(label)
                yield result_bar
        else:
            # Show radio buttons to vote
            yield Static("Cast your vote:", classes="poll-header")
            with RadioSet(id="poll-options"):
                for option in self.poll["options"]:
                    yield RadioButton(option["title"])

        # Add the footer with expiry and total votes
        with Horizontal(classes="poll-footer"):
            total_votes = self.poll.get("votes_count", 0)
            yield Static(f"{total_votes} votes", classes="poll-total-votes")

            expires_at = self.poll.get("expires_at")
            if expires_at:
                expiry_str = format_datetime(expires_at)
                status_text = "Expired" if self.poll.get("expired") else "Expires"
                yield Static(f"{status_text}: {expiry_str}", classes="poll-expiry")

    @on(RadioSet.Changed)
    def on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        """Handle a vote."""
        if event.radio_set.pressed_index is not None:
            # Check if already voted, just in case
            if self.poll.get("voted"):
                return
            self.post_message(
                VoteOnPoll(
                    self.poll["id"],
                    event.radio_set.pressed_index,
                    self.timeline_id,
                    self.post_id,
                )
            )
            # Disable the radio set to prevent re-voting
            event.radio_set.disabled = True


class Post(Vertical):
    """A widget to display a single post."""

    def __init__(self, post, timeline_id: str, **kwargs):
        super().__init__(**kwargs)
        self.post = post
        self.timeline_id = timeline_id
        self.add_class("timeline-item")
        status_to_display = self.post.get("reblog") or self.post
        self.created_at_str = format_datetime(status_to_display["created_at"])

    def on_mount(self):
        status_to_display = self.post.get("reblog") or self.post
        if status_to_display.get("favourited"):
            self.add_class("favourited")
        if status_to_display.get("reblogged"):
            self.add_class("reblogged")

        # Check for mentions
        try:
            user_id = self.app.me["id"]
            if any(
                mention["id"] == user_id
                for mention in status_to_display.get("mentions", [])
            ):
                self.add_class("mention")
        except Exception as e:
            log.error(f"Could not check for mentions: {e}")

    def compose(self):
        reblog = self.post.get("reblog")
        is_reblog = reblog is not None
        status_to_display = reblog or self.post

        if is_reblog:
            booster_display_name = self.post["account"]["display_name"]
            booster_acct = self.post["account"]["acct"]
            yield Static(
                f"🚀 Boosted by {booster_display_name} (@{booster_acct})",
                classes="boost-header",
            )

        spoiler_text = status_to_display.get("spoiler_text")
        author_display_name = status_to_display["account"]["display_name"]
        author_acct = status_to_display["account"]["acct"]
        author = f"{author_display_name} (@{author_acct})"
        self.border_title = author
        if spoiler_text:
            self.border_title = spoiler_text
            self.border_subtitle = author

        yield Markdown(get_full_content_md(status_to_display), open_links=False)

        if status_to_display.get("poll"):
            yield PollWidget(
                status_to_display["poll"],
                timeline_id=self.timeline_id,
                post_id=status_to_display["id"],
            )

        if self.app.config.image_support and status_to_display.get("media_attachments"):
            for media in status_to_display["media_attachments"]:
                if media["type"] == "image":
                    yield ImageWidget(media["url"], self.app.config)

        with Horizontal(classes="post-footer"):
            yield LoadingIndicator(classes="action-spinner")
            yield Static(
                f"Boosts: {status_to_display.get('reblogs_count', 0)}", id="boost-count"
            )
            yield Static(
                f"Likes: {status_to_display.get('favourites_count', 0)}",
                id="like-count",
            )
            yield Static(format_datetime(status_to_display["created_at"]), classes="timestamp")

            visibility_icons = {
                "public": "🌐",
                "unlisted": "🔑",
                "private": "👥",
                "direct": "🔒",
            }
            visibility = status_to_display.get("visibility")
            vis_icon = visibility_icons.get(visibility, "")
            yield Static(vis_icon, classes="visibility-icon")

    def show_spinner(self):
        self.query_one(".action-spinner").display = True

    def hide_spinner(self):
        self.query_one(".action-spinner").display = False

    def update_from_post(self, post):
        self.post = post
        status_to_display = self.post.get("reblog") or self.post

        # Update classes
        self.remove_class("favourited", "reblogged")
        if status_to_display.get("favourited"):
            self.add_class("favourited")
        if status_to_display.get("reblogged"):
            self.add_class("reblogged")

        # Update stats
        self.query_one("#boost-count").update(
            f"Boosts: {status_to_display.get('reblogs_count', 0)}"
        )
        self.query_one("#like-count").update(
            f"Likes: {status_to_display.get('favourites_count', 0)}"
        )
        self.hide_spinner()

        # Force re-render of the content
        for md in self.query(Markdown):
            md.remove()
        self.mount(Markdown(get_full_content_md(status_to_display), open_links=False), before=self.query_one(".post-footer"))

        # Re-render the poll if it exists
        for poll_widget in self.query(PollWidget):
            poll_widget.remove()
        if status_to_display.get("poll"):
            self.mount(
                PollWidget(
                    status_to_display["poll"],
                    timeline_id=self.timeline_id,
                    post_id=status_to_display["id"],
                ),
                after=self.query_one(".post-footer"),
            )

    def on_click(self, event: events.Click) -> None:
        event.stop()
        self.post_message(SelectPost(self))

    @on(Markdown.LinkClicked)
    def on_markdown_link_clicked(self, event: Markdown.LinkClicked) -> None:
        """Handle a link being clicked in the Markdown."""
        event.stop()
        href = event.href
        # Heuristic to check if the link is a hashtag
        if "/tags/" in href:
            hashtag = href.split("/tags/")[-1].rstrip("/")
            self.post_message(ViewHashtag(hashtag))
            return  # Explicitly stop processing here

        # If it's not a hashtag, open it in the browser
        self.app.action_link_clicked(href)

    def get_created_at(self) -> datetime | None:
        status = self.post.get("reblog") or self.post
        if status and "created_at" in status:
            ts = status["created_at"]
            if isinstance(ts, datetime):
                return ts
            return datetime.fromisoformat(ts.replace("Z", "+00:00"))
        return None


class GapIndicator(Widget):
    """A widget to indicate a gap in the timeline."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_class("gap-indicator")

    def compose(self):
        yield Static("...")


class Notification(Widget):
    """A widget to display a single notification."""

    def __init__(self, notif, **kwargs):
        super().__init__(**kwargs)
        self.notif = notif
        self.add_class("timeline-item")

        created_at = None
        if self.notif["type"] == "mention":
            created_at = self.notif["status"]["created_at"]
        else:
            created_at = self.notif["created_at"]
        self.created_at_str = format_datetime(created_at)

    def compose(self):
        notif_type = self.notif["type"]
        author = self.notif["account"]
        author_display_name = author["display_name"]
        author_acct = f"@{author['acct']}"
        author_str = f"{author_display_name} ({author_acct})"

        created_at = None
        if self.notif["type"] == "mention":
            created_at = self.notif["status"]["created_at"]
        else:
            created_at = self.notif["created_at"]

        if notif_type == "mention":
            status = self.notif["status"]
            spoiler_text = status.get("spoiler_text")

            self.border_title = f"Mention from {author_str}"
            if spoiler_text:
                self.border_title = spoiler_text
                self.border_subtitle = f"Mention from {author_str}"

            yield Markdown(get_full_content_md(status), open_links=False)
            if self.app.config.image_support and status.get("media_attachments"):
                for media in status["media_attachments"]:
                    if media["type"] == "image":
                        yield ImageWidget(media["url"], self.app.config)
            with Horizontal(classes="post-footer"):
                yield Static(
                    f"Boosts: {status.get('reblogs_count', 0)}", id="boost-count"
                )
                yield Static(
                    f"Likes: {status.get('favourites_count', 0)}", id="like-count"
                )
                yield Static(format_datetime(created_at), classes="timestamp")

        elif notif_type == "favourite":
            status = self.notif["status"]
            self.border_title = f"❤️ {author_str} favourited your post:"
            yield Markdown(get_full_content_md(status), open_links=False)
            if self.app.config.image_support and status.get("media_attachments"):
                for media in status["media_attachments"]:
                    if media["type"] == "image":
                        yield ImageWidget(media["url"], self.app.config)
            with Horizontal(classes="post-footer"):
                yield Static(format_datetime(created_at), classes="timestamp")

        elif notif_type == "reblog":
            status = self.notif["status"]
            self.border_title = f"🚀 {author_str} boosted your post:"
            yield Markdown(get_full_content_md(status), open_links=False)
            if self.app.config.image_support and status.get("media_attachments"):
                for media in status["media_attachments"]:
                    if media["type"] == "image":
                        yield ImageWidget(media["url"], self.app.config)
            with Horizontal(classes="post-footer"):
                yield Static(format_datetime(created_at), classes="timestamp")

        elif notif_type == "follow":
            self.border_title = f"👋 {author_str} followed you."
            with Horizontal(classes="post-footer"):
                yield Static(format_datetime(created_at), classes="timestamp")

        elif notif_type == "poll":
            status = self.notif["status"]
            self.border_title = "📊 A poll you participated in has ended:"
            yield PollWidget(
                status["poll"], timeline_id="notifications", post_id=status["id"]
            )
            with Horizontal(classes="post-footer"):
                yield Static(format_datetime(created_at), classes="timestamp")
        elif notif_type == "update":
            status = self.notif["status"]
            self.border_title = (
                f"✍️ A post by {author_str} you interacted with was updated"
            )
            self.border_subtitle = f"{author_str}"
            yield Markdown(get_full_content_md(status))
            if self.app.config.image_support and status.get("media_attachments"):
                for media in status["media_attachments"]:
                    if media["type"] == "image":
                        yield ImageWidget(media["url"], self.app.config)
            with Horizontal(classes="post-footer"):
                yield Static(format_datetime(created_at), classes="timestamp")

        else:
            yield Static(f"Unsupported notification type: {notif_type}")

    def on_click(self, event: events.Click) -> None:
        event.stop()
        self.post_message(SelectPost(self))

    @on(Markdown.LinkClicked)
    def on_markdown_link_clicked(self, event: Markdown.LinkClicked) -> None:
        """Handle a link being clicked in the Markdown."""
        event.stop()
        href = event.href
        # Heuristic to check if the link is a hashtag
        if "/tags/" in href:
            hashtag = href.split("/tags/")[-1].rstrip("/")
            self.post_message(ViewHashtag(hashtag))
            return  # Explicitly stop processing here

        # If it's not a hashtag, open it in the browser
        self.app.action_link_clicked(href)

    def get_created_at(self) -> datetime | None:
        ts_str = None
        if self.notif["type"] == "mention":
            ts_str = self.notif.get("status", {}).get("created_at")
        else:
            ts_str = self.notif.get("created_at")

        if ts_str:
            if isinstance(ts_str, datetime):
                return ts_str
            return datetime.fromisoformat(ts_str.replace("Z", "+00:00"))
        return None


class SearchResult(Widget, can_focus=True):
    """A base class for search results."""

    pass


class AccountResult(SearchResult):
    """A widget to display an account search result."""

    def __init__(self, account: dict, **kwargs):
        super().__init__(**kwargs)
        self.account = account
        self.add_class("search-result")

    def compose(self):
        with Vertical():
            yield Static(
                f"[bold]{self.account['display_name']}[/bold] @{self.account['acct']}"
            )
            yield Markdown(to_markdown(self.account["note"]), open_links=False)


class HashtagResult(SearchResult):
    """A widget to display a hashtag search result."""

    def __init__(self, hashtag: dict, **kwargs):
        super().__init__(**kwargs)
        self.hashtag = hashtag
        self.add_class("search-result")

    def compose(self):
        yield Static(f"#{self.hashtag['name']}")
        # You can add more info here, like recent usage stats if available


class StatusResult(SearchResult):
    """A widget to display a status search result."""

    def __init__(self, status: dict, **kwargs):
        super().__init__(**kwargs)
        self.status = status
        self.add_class("search-result")

    def compose(self):
        yield Post(self.status, timeline_id="search")


class ConversationSummary(Widget, can_focus=True):
    """A widget to display a conversation summary."""

    def __init__(self, conversation: dict, **kwargs):
        super().__init__(**kwargs)
        self.conversation = conversation
        self.add_class("timeline-item")
        if conversation.get("unread"):
            self.add_class("unread")

    def on_click(self, event: events.Click) -> None:
        event.stop()
        self.post_message(SelectPost(self))

    def compose(self):
        accounts = self.conversation.get("accounts", [])
        last_status = self.conversation.get("last_status")
        is_unread = self.conversation.get("unread")

        # Filter out the current user's account to display others
        other_accounts = [
            acc for acc in accounts if acc["id"] != self.app.me["id"]
        ]
        
        participant_names = ", ".join(
            [f"@{acc['acct']}" for acc in other_accounts]
        )
        
        icon = "📩" if is_unread else "📭"
        self.border_title = f"{icon} DM with {participant_names}"

        if last_status:
            snippet = to_markdown(last_status.get("content", ""))
            # Truncate snippet to a reasonable length
            if len(snippet) > 100:
                snippet = snippet[:97] + "..."
            yield Markdown(snippet, open_links=False)
            
            timestamp = format_datetime(last_status.get("created_at"))
            with Horizontal(classes="post-footer"):
                yield Static(format_datetime(last_status.get("created_at")), classes="timestamp")
        else:
            yield Static("No messages yet.")



from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Button, Label, Input, Static, TextArea, Select, Header
from textual.containers import Grid, Horizontal, Vertical, VerticalScroll
from textual import on
from mastui.utils import LANGUAGE_OPTIONS, html_to_plain_text

class EditPostScreen(ModalScreen):
    """A modal screen for editing a post."""

    BINDINGS = [
        ("escape", "app.pop_screen", "Cancel Edit"),
    ]

    def __init__(self, status: dict, max_characters: int = 500, **kwargs):
        super().__init__(**kwargs)
        self.status = status
        self.max_characters = max_characters

    def compose(self):
        with Vertical(id="post_dialog") as d:
            d.border_title = "Edit Post"
            with VerticalScroll(id="post_content_container"):
                yield TextArea(html_to_plain_text(self.status.get('content', '')), id="post_content", language="markdown")
                with Horizontal(id="post_options"):
                    yield Label("CW:", classes="post_option_label")
                    yield Input(
                        value=self.status.get('spoiler_text', ''),
                        placeholder="Content warning", 
                        id="cw_input"
                    )
                with Horizontal(id="post_language_container"):
                    yield Label("Language:", classes="post_option_label")
                    yield Select(
                        LANGUAGE_OPTIONS, 
                        value=self.status.get('language', 'en'), 
                        id="language_select"
                    )
            with Horizontal(id="post_buttons"):
                yield Label(f"{self.max_characters}", id="character_limit")
                yield Button("Save", variant="primary", id="save_button")
                yield Button("Cancel", id="cancel_button")

    def on_mount(self):
        self.query_one("#post_content").focus()
        self.update_character_limit()

    @on(Input.Changed)
    @on(TextArea.Changed)
    def update_character_limit(self):
        """Updates the character limit."""
        content_len = len(self.query_one("#post_content").text)
        cw_len = len(self.query_one("#cw_input").value)
        remaining = self.max_characters - content_len - cw_len
        
        limit_label = self.query_one("#character_limit")
        limit_label.update(f"{remaining}")
        limit_label.set_class(remaining < 0, "character-limit-error")

    @on(Button.Pressed)
    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "save_button":
            content = self.query_one("#post_content").text
            spoiler_text = self.query_one("#cw_input").value
            language = self.query_one("#language_select").value
            
            result = {
                "content": content,
                "spoiler_text": spoiler_text,
                "language": language,
            }
            self.dismiss(result)
        elif event.button.id == "cancel_button":
            self.dismiss(None)



from mastodon import Mastodon, MastodonError
from requests import Session
import logging

log = logging.getLogger(__name__)


def get_api(config_obj):
    """Initializes and returns a Mastodon API instance."""
    conf = config_obj
    if conf.mastodon_access_token:
        s = Session()
        s.verify = conf.ssl_verify
        return Mastodon(
            access_token=conf.mastodon_access_token,
            api_base_url=f"https://{conf.mastodon_host}",
            session=s,
            mastodon_version="4.0.0",
        )
    return None


def login(host, client_id, client_secret, auth_code, ssl_verify=True):
    """Logs in to a Mastodon instance using an auth code and returns the API object or an error."""
    try:
        s = Session()
        s.verify = ssl_verify
        mastodon = Mastodon(
            client_id=client_id,
            client_secret=client_secret,
            api_base_url=f"https://{host}",
            session=s,
            mastodon_version="4.0.0",
        )
        access_token = mastodon.log_in(
            code=auth_code,
            redirect_uri="urn:ietf:wg:oauth:2.0:oob",
            scopes=["read", "write", "follow", "push"],
        )
        # Create the content for the .env file
        env_content = (
            f"MASTODON_HOST={host}\n"
            f"MASTODON_CLIENT_ID={client_id}\n"
            f"MASTODON_CLIENT_SECRET={client_secret}\n"
            f"MASTODON_ACCESS_TOKEN={access_token}\n"
        )

        final_session = Session()
        final_session.verify = ssl_verify
        # Re-initialize with the new access token
        api = Mastodon(
            access_token=access_token,
            api_base_url=f"https://{host}",
            session=final_session,
        )
        return api, env_content, None
    except MastodonError as e:
        log.error(f"Mastodon API error during login for host '{host}': {e}", exc_info=True)
        return None, None, str(e)
    except Exception as e:
        log.error(f"Unexpected error during login for host '{host}': {e}", exc_info=True)
        return None, None, str(e)


def create_app(host, ssl_verify=True):
    """Creates a new Mastodon app and returns the auth URL."""
    try:
        s = Session()
        s.verify = ssl_verify
        client_id, client_secret = Mastodon.create_app(
            "mastui",
            api_base_url=f"https://{host}",
            scopes=["read", "write", "follow", "push"],
            redirect_uris="urn:ietf:wg:oauth:2.0:oob",
            session=s,
        )

        mastodon = Mastodon(
            client_id=client_id,
            client_secret=client_secret,
            api_base_url=f"https://{host}",
            session=s,
            mastodon_version="4.0.0",
        )
        auth_url = mastodon.auth_request_url(
            redirect_uris="urn:ietf:wg:oauth:2.0:oob",
            scopes=["read", "write", "follow", "push"],
        )
        return auth_url, client_id, client_secret, None

    except MastodonError as e:
        log.error(
            f"Mastodon API error during app creation for host '{host}': {e}",
            exc_info=True,
        )

        # Log the raw response if available on the exception
        if hasattr(e, "text") and e.text:
            log.debug(f"Raw server response from '{host}':\n{e.text}")

        error_message = str(e)
        if "Expecting value" in error_message:
            error_message = f"Could not parse server response from '{host}'. The instance may be offline, misconfigured, or not a Mastodon instance."

        return None, None, None, error_message
    except Exception as e:
        # Catch any other potential errors (e.g., network issues not caught by MastodonError)
        log.error(
            f"Unexpected error during app creation for host '{host}': {e}",
            exc_info=True,
        )
        return None, None, None, str(e)



import clipman
from textual.app import App, ComposeResult
from textual.containers import Grid, Vertical
from textual.widgets import Button, Label, Input, Static, TextArea, LoadingIndicator, Header, ContentSwitcher
from textual.screen import ModalScreen
from urllib.parse import urlparse
from rich.panel import Panel

from mastui.mastodon_api import login, create_app
import logging

log = logging.getLogger(__name__)


class LoginScreen(ModalScreen):
    """Screen for user to login."""

    def __init__(self, host: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.host = host
        self.client_id = None
        self.client_secret = None

    def compose(self) -> ComposeResult:
        with Vertical(id="login-dialog") as d:
            d.border_title = "Mastui Login"
            with ContentSwitcher(initial="login-initial-view"):
                with Vertical(id="login-initial-view"):
                    yield Static("[bold]Welcome to Mastui![/bold]\n\nEnter your Mastodon instance to get started.")
                    with Grid(id="login-grid"):
                        yield Label("Mastodon Instance:")
                        yield Input(placeholder="mastodon.social", id="host", value=self.host or "")
                    yield Static() # Spacer
                    yield Button("Get Auth Link", variant="primary", id="get_auth")

                with Vertical(id="login-auth-view"):
                    yield Static("1. An authorization link has been copied to your clipboard (if possible).")
                    yield Static("   Open it in your browser to grant Mastui access.")
                    yield TextArea("", id="auth_link", read_only=True)
                    yield Static("\n2. Paste the authorization code you received here:")
                    yield Input(placeholder="Authorization Code", id="auth_code")
                    yield Button("Login", variant="primary", id="login")

                with Vertical(id="login-loading-view", classes="centered"):
                     yield Static("Working...")
                     yield LoadingIndicator()

            yield Static(id="login-status")


    def on_mount(self) -> None:
        """Called when the screen is mounted."""
        host_input = self.query_one("#host")
        if self.host:
            host_input.disabled = True
            self.query_one("#get_auth").focus()
        else:
            host_input.focus()

    def clean_host(self, host_input: str) -> str:
        """Cleans the host input to be a valid domain."""
        if not host_input:
            return ""
        
        # Prepend a scheme if one isn't present, for urlparse to work correctly.
        if not host_input.startswith(('http://', 'https ')):
            host_input = 'https://' + host_input
            
        parsed_url = urlparse(host_input)
        # netloc contains the domain and potentially the port
        return parsed_url.netloc

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        status = self.query_one("#login-status")
        switcher = self.query_one(ContentSwitcher)

        if event.button.id == "get_auth":
            host_input_widget = self.query_one("#host")
            host_value = host_input_widget.value
            if not host_value:
                status.update("Please enter a Mastodon instance.")
                return

            cleaned_host = self.clean_host(host_value)
            if not cleaned_host:
                status.update("Invalid instance name.")
                return
            
            host_input_widget.value = cleaned_host
            switcher.current = "login-loading-view"
            status.update("")
            self.run_worker(
                lambda: (
                    result := create_app(cleaned_host, self.app.ssl_verify),
                    log.debug(f"create_app() returned: {result}"),
                    self.app.call_from_thread(self.on_auth_link_created, result),
                ),
                exclusive=True,
                thread=True,
            )

        elif event.button.id == "login":
            auth_code = self.query_one("#auth_code").value
            host = self.query_one("#host").value
            if not auth_code:
                status.update("Please enter the authorization code.")
                return

            switcher.current = "login-loading-view"
            status.update("")
            self.run_worker(
                lambda: (
                    result := login(
                        host,
                        self.client_id,
                        self.client_secret,
                        auth_code,
                        self.app.ssl_verify,
                    ),
                    log.debug(f"login() returned: {result}"),
                    self.app.call_from_thread(self.on_login_complete, result),
                ),
                exclusive=True,
                thread=True,
            )

    def on_auth_link_created(self, result) -> None:
        """Callback for when the auth link is created."""
        status = self.query_one("#login-status")
        switcher = self.query_one(ContentSwitcher)
        
        try:
            auth_url, client_id, client_secret, error = result
            self.client_id = client_id
            self.client_secret = client_secret

            if error:
                status.update(f"Error: {error}")
                switcher.current = "login-initial-view"
                return

            try:
                clipman.init()
                clipman.set(auth_url)
            except clipman.exceptions.ClipmanBaseException as e:
                log.warning(f"Could not copy to clipboard: {e}")

            auth_link_input = self.query_one("#auth_link")
            auth_link_input.text = auth_url
            switcher.current = "login-auth-view"
            self.query_one("#auth_code").focus()
        except (TypeError, ValueError) as e:
            log.error(f"Error unpacking result in on_auth_link_created: {result} - {e}", exc_info=True)
            status.update("An unexpected error occurred. See log for details.")
            switcher.current = "login-initial-view"
        except Exception as e:
            log.error(f"Unexpected error in on_auth_link_created: {e}", exc_info=True)
            status.update("An unexpected error occurred. See log for details.")
            switcher.current = "login-initial-view"

    def on_login_complete(self, result) -> None:
        """Callback for when the login is complete."""
        log.debug(f"on_login_complete received result: {result}")
        status = self.query_one("#login-status")
        switcher = self.query_one(ContentSwitcher)

        api, env_content, error = result

        if api:
            self.dismiss((api, env_content))
            return
        else:
            status.update(f"Login failed: {error}")
            switcher.current = "login-auth-view"
